import logging

import pytest
from traitlets import HasTraits, TraitError

from jupyter_events.traits import Handlers


class HasHandlers(HasTraits):
    handlers = Handlers(None, allow_none=True)


def test_good_handlers_value():
    handlers = [logging.NullHandler(), logging.NullHandler()]
    obj = HasHandlers(handlers=handlers)
    assert obj.handlers == handlers


def test_bad_handlers_values():
    handlers = [0, 1]

    with pytest.raises(TraitError):
        HasHandlers(handlers=handlers)


def test_mixed_handlers_values():
    handlers = [logging.NullHandler(), 1]
    with pytest.raises(TraitError):
        HasHandlers(handlers=handlers)



"""The cli for jupyter events."""
import json
import pathlib
import platform

import click
from jsonschema import ValidationError
from rich.console import Console
from rich.json import JSON
from rich.markup import escape
from rich.padding import Padding
from rich.style import Style

from jupyter_events.schema import EventSchema, EventSchemaFileAbsent, EventSchemaLoadingError

WIN = platform.system() == "Windows"


class RC:
    """Return code enum."""

    OK = 0
    INVALID = 1
    UNPARSABLE = 2
    NOT_FOUND = 3


class EMOJI:
    """Terminal emoji enum"""

    X = "XX" if WIN else "\u274c"
    OK = "OK" if WIN else "\u2714"


console = Console()
error_console = Console(stderr=True)


@click.group()
@click.version_option()
def main() -> None:
    """A simple CLI tool to quickly validate JSON schemas against
    Jupyter Event's custom validator.

    You can see Jupyter Event's meta-schema here:

        https://raw.githubusercontent.com/jupyter/jupyter_events/main/jupyter_events/schemas/event-metaschema.yml
    """
    pass


@click.command()
@click.argument("schema")
@click.pass_context
def validate(ctx: click.Context, schema: str) -> int:
    """Validate a SCHEMA against Jupyter Event's meta schema.

    SCHEMA can be a JSON/YAML string or filepath to a schema.
    """
    console.rule("Validating the following schema", style=Style(color="blue"))

    _schema = None
    try:
        # attempt to read schema as a serialized string
        _schema = EventSchema._load_schema(schema)
    except EventSchemaLoadingError:
        # pass here to avoid printing traceback of this exception if next block
        # excepts
        pass

    # if not a serialized schema string, try to interpret it as a path to schema file
    if _schema is None:
        schema_path = pathlib.Path(schema)
        try:
            _schema = EventSchema._load_schema(schema_path)
        except (EventSchemaLoadingError, EventSchemaFileAbsent) as e:
            # no need for full tracestack for user error exceptions. just print
            # the error message and return
            error_console.print(f"[bold red]ERROR[/]: {e}")
            return ctx.exit(RC.UNPARSABLE)

    # Print what was found.
    schema_json = JSON(json.dumps(_schema))
    console.print(Padding(schema_json, (1, 0, 1, 4)))
    # Now validate this schema against the meta-schema.
    try:
        EventSchema(_schema)
        console.rule("Results", style=Style(color="green"))
        out = Padding(f"[green]{EMOJI.OK}[white] Nice work! This schema is valid.", (1, 0, 1, 0))
        console.print(out)
        return ctx.exit(RC.OK)
    except ValidationError as err:
        error_console.rule("Results", style=Style(color="red"))
        error_console.print(f"[red]{EMOJI.X} [white]The schema failed to validate.")
        error_console.print("\nWe found the following error with your schema:")
        out = escape(str(err))  # type:ignore[assignment]
        error_console.print(Padding(out, (1, 0, 1, 4)))
        return ctx.exit(RC.INVALID)


main.add_command(validate)



"""Yaml utilities."""
from __future__ import annotations

from pathlib import Path, PurePath
from typing import Any

from yaml import dump as ydump
from yaml import load as yload

try:
    from yaml import CSafeDumper as SafeDumper
    from yaml import CSafeLoader as SafeLoader
except ImportError:  # pragma: no cover
    from yaml import SafeDumper, SafeLoader  # type:ignore[assignment]


def loads(stream: Any) -> Any:
    """Load yaml from a stream."""
    return yload(stream, Loader=SafeLoader)


def dumps(stream: Any) -> str:
    """Parse the first YAML document in a stream as an object."""
    return ydump(stream, Dumper=SafeDumper)


def load(fpath: str | PurePath) -> Any:
    """Load yaml from a file."""
    # coerce PurePath into Path, then read its contents
    data = Path(str(fpath)).read_text(encoding="utf-8")
    return loads(data)


def dump(data: Any, outpath: str | PurePath) -> None:
    """Parse the a YAML document in a file as an object."""
    Path(outpath).write_text(dumps(data), encoding="utf-8")



"""Trait types for events."""
from __future__ import annotations

import logging
import typing as t

from traitlets import TraitError, TraitType

baseclass = TraitType
if t.TYPE_CHECKING:
    baseclass = TraitType[t.Any, t.Any]  # type:ignore[misc]


class Handlers(baseclass):  # type:ignore[type-arg]
    """A trait that takes a list of logging handlers and converts
    it to a callable that returns that list (thus, making this
    trait pickleable).
    """

    info_text = "a list of logging handlers"

    def validate_elements(self, obj: t.Any, value: t.Any) -> None:
        """Validate the elements of an object."""
        if len(value) > 0:
            # Check that all elements are logging handlers.
            for el in value:
                if isinstance(el, logging.Handler) is False:
                    self.element_error(obj)

    def element_error(self, obj: t.Any) -> None:
        """Raise an error for bad elements."""
        msg = f"Elements in the '{self.name}' trait of an {obj.__class__.__name__} instance must be Python `logging` handler instances."
        raise TraitError(msg)

    def validate(self, obj: t.Any, value: t.Any) -> t.Any:
        """Validate an object."""
        # If given a callable, call it and set the
        # value of this trait to the returned list.
        # Verify that the callable returns a list
        # of logging handler instances.
        if callable(value):
            out = value()
            self.validate_elements(obj, out)
            return out
        # If a list, check it's elements to verify
        # that each element is a logging handler instance.
        elif isinstance(value, list):
            self.validate_elements(obj, value)
            return value
        else:
            self.error(obj, value)  # type:ignore[no-untyped-call]
